import pytest

from anaplan_sdk import Client
from anaplan_sdk.exceptions import (
    AnaplanException,
    InvalidCredentialsException,
    InvalidIdentifierException,
)
from anaplan_sdk.models import Model, TaskStatus, TaskSummary, Workspace


def test_list_workspaces(client):
    workspaces, search = client.get_workspaces(), client.get_workspaces("Demo")
    assert isinstance(workspaces, list)
    assert all(isinstance(workspace, Workspace) for workspace in workspaces)
    assert all(isinstance(workspace, Workspace) for workspace in search)
    assert len(workspaces) > 0
    assert len(search) > 0
    assert len(search) < len(workspaces)


def test_broken_list_files_raises_invalid_identifier_error(broken_client):
    with pytest.raises(InvalidIdentifierException):
        broken_client.get_files()


def unauthenticated_client_raises_exception():
    with pytest.raises(InvalidCredentialsException):
        _ = Client(user_email="invalid_email", password="pass")


def test_broken_client_alm_raises(broken_client: Client):
    with pytest.raises(ValueError):
        _ = broken_client.alm


def test_broken_client_transactional_raises(broken_client: Client):
    with pytest.raises(ValueError):
        _ = broken_client.tr


def test_file_creation_raises_exception(client: Client):
    with pytest.raises(InvalidIdentifierException):
        client.upload_file(115000000000, b"")


def test_list_models(client: Client):
    models = client.get_models()
    current_only = client.get_models(True)
    search = client.get_models(search_pattern="Demo")
    assert isinstance(models, list)
    assert all(isinstance(model, Model) for model in models)
    assert all(isinstance(model, Model) for model in current_only)
    assert all(isinstance(model, Model) for model in search)
    assert len(models) > len(current_only) > len(search) > 0


def test_list_models_multi_page(client_small_pages: Client):
    models = client_small_pages.get_models()
    assert isinstance(models, list)
    assert len(models) > 0
    assert all(isinstance(model, Model) for model in models)
    assert len(models) == len(set(m.id for m in models))  # Ensure no duplicates when paginating


def test_list_actions(client: Client):
    actions = client.get_actions()
    assert isinstance(actions, list)
    assert len(actions) > 0


def test_list_files(client: Client):
    files = client.get_files()
    assert isinstance(files, list)
    assert len(files) > 0


def test_list_processes(client: Client):
    processes = client.get_processes()
    assert isinstance(processes, list)
    assert len(processes) > 0


def test_list_imports(client: Client):
    imports = client.get_imports()
    assert isinstance(imports, list)
    assert len(imports) > 0


def test_list_exports(client: Client):
    exports = client.get_exports()
    assert isinstance(exports, list)
    assert len(exports) > 0


def test_upload_file_stream(client: Client, test_file):
    client.upload_file_stream(test_file, (str(i) for i in range(10)))
    out = client.get_file(test_file)
    assert out == b"0123456789"


def test_get_file_stream(client: Client, test_file):
    for chunk in client.get_file_stream(test_file):
        assert isinstance(chunk, bytes)


def test_upload_and_download_file(client: Client, test_file):
    client.upload_file(test_file, "Hi!")
    out = client.get_file(test_file)
    assert out == b"Hi!"


def test_run_process(client: Client, test_action):
    client.run_action(test_action)


def test_list_task_statuses(client: Client, test_action):
    task_statuses = client.get_task_summaries(test_action)
    assert isinstance(task_statuses, list)
    assert all(isinstance(status, TaskSummary) for status in task_statuses)
    assert len(task_statuses) > 0


def test_get_task_status(client: Client, test_action):
    task_status = client.get_task_status(test_action, client.run_action(test_action, False).id)
    assert isinstance(task_status, TaskStatus)


def test_invalid_file_id_raises_exception(client: Client):
    with pytest.raises(InvalidIdentifierException):
        client.get_file(1)


def test_run_nonexistent_action_raises_exception(client: Client):
    with pytest.raises(InvalidIdentifierException):
        client.run_action(1)


def test_upload_empty_file(client: Client, test_file):
    with pytest.raises(AnaplanException):
        client.upload_file(test_file, b"")
        # Error occurs here, since the file does not actually exist
        client.get_file(test_file)


def test_with_model(client: Client):
    other_ws_id = other_model_id = "123"
    other_client = client.with_model(other_model_id, other_ws_id)
    assert isinstance(other_client, Client)
    assert other_client._workspace_id == other_ws_id
    assert other_client._model_id == other_model_id
    assert other_client.alm._model_id == other_model_id
    assert other_client.tr._model_id == other_model_id
    assert other_client.alm._model_id == other_model_id



import logging
import string
import sys
from os import getenv
from random import choices
from typing import Any

import pytest

from anaplan_sdk.models.cloud_works import (
    AnaplanTarget,
    AzureBlobConnectionInput,
    ConnectionInput,
    FileSource,
    IntegrationInput,
    IntegrationJobInput,
    IntegrationProcessInput,
    NotificationConfigInput,
    NotificationInput,
    NotificationItemInput,
    ScheduleInput,
)
from anaplan_sdk.models.flows import FlowInput, FlowStepInput

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s] %(message)s")
logging.getLogger("httpx").setLevel(logging.ERROR)
logging.getLogger("anaplan_sdk").setLevel(logging.INFO)


@pytest.fixture(scope="session")
def py_version():
    return f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"


@pytest.fixture(scope="session")
def list_items_long():
    return [{"name": i, "code": i} for i in range(200_000)]  # Force several batches


@pytest.fixture(scope="session")
def list_items_short():
    return [{"name": i, "code": i} for i in range(1_000)]  # Single batch


@pytest.fixture(scope="session")
def connection_id():
    return "8b2d5f3a2ff64f13ab52e5b993896386"


@pytest.fixture(scope="session")
def error_run_id():
    return "910a68fb814e4225ad683bdafb70ae65"


@pytest.fixture(scope="session")
def scim_user_id():
    return "38a0546fd5894c1fac87f8fb71566b3f"


@pytest.fixture
def name():
    return "Test_" + "".join(choices(string.ascii_uppercase + string.digits, k=12))


@pytest.fixture
def az_blob_connection(name):
    return ConnectionInput(
        type="AzureBlob",
        body=AzureBlobConnectionInput(
            storage_account_name=getenv("AZ_STORAGE_ACCOUNT"),
            container_name="raw",
            name=name,
            sas_token=getenv("AZ_STORAGE_SAS_TOKEN"),
        ),
    )


@pytest.fixture
def az_blob_connection_dict(name):
    return {
        "type": "AzureBlob",
        "body": {
            "storageAccountName": getenv("AZ_STORAGE_ACCOUNT"),
            "containerName": "raw",
            "name": name,
            "sasToken": getenv("AZ_STORAGE_SAS_TOKEN"),
        },
    }


@pytest.fixture
def integration_pydantic(name, connection_id):
    source = FileSource(
        type="AzureBlob", connection_id=connection_id, file="dummy/liquor_sales.csv"
    )
    target = AnaplanTarget(action_id=112000000064, file_id=113000000055)
    return IntegrationInput(
        name=name,
        workspace_id=getenv("ANAPLAN_SDK_TEST_WORKSPACE_ID"),
        model_id=getenv("ANAPLAN_SDK_TEST_MODEL_ID"),
        jobs=[IntegrationJobInput(type="AzureBlobToAnaplan", sources=[source], targets=[target])],
    )


@pytest.fixture
def multi_step_integration_pydantic(name, connection_id):
    source = FileSource(
        type="AzureBlob", connection_id=connection_id, file="dummy/liquor_sales.csv"
    )
    target = AnaplanTarget(action_id=112000000064, file_id=113000000055)
    job = IntegrationJobInput(type="AzureBlobToAnaplan", sources=[source], targets=[target])
    return IntegrationInput(
        name=name,
        process_id=118000000012,
        workspace_id=getenv("ANAPLAN_SDK_TEST_WORKSPACE_ID"),
        model_id=getenv("ANAPLAN_SDK_TEST_MODEL_ID"),
        jobs=[job, job],
    )


@pytest.fixture
def process_integration_pydantic(name):
    return IntegrationProcessInput(
        name=name,
        process_id=118000000012,
        workspace_id=getenv("ANAPLAN_SDK_TEST_WORKSPACE_ID"),
        model_id=getenv("ANAPLAN_SDK_TEST_MODEL_ID"),
    )


@pytest.fixture
def integration_dict(name, connection_id):
    return {
        "name": name,
        "version": "2.0",
        "workspaceId": getenv("ANAPLAN_SDK_TEST_WORKSPACE_ID"),
        "modelId": getenv("ANAPLAN_SDK_TEST_MODEL_ID"),
        "nuxVisible": False,
        "jobs": [
            {
                "type": "AzureBlobToAnaplan",
                "sources": [
                    {
                        "connectionId": connection_id,
                        "type": "AzureBlob",
                        "file": "dummy/liquor_sales.csv",
                    }
                ],
                "targets": [
                    {"type": "Anaplan", "actionId": "112000000001", "fileId": "112000000001"}
                ],
            }
        ],
    }


@pytest.fixture
def multi_step_integration_dict(name, connection_id) -> dict[str, Any]:
    return {
        "name": name,
        "version": "2.0",
        "workspaceId": getenv("ANAPLAN_SDK_TEST_WORKSPACE_ID"),
        "modelId": getenv("ANAPLAN_SDK_TEST_MODEL_ID"),
        "processId": "118000000012",
        "nuxVisible": False,
        "jobs": [
            {
                "type": "AzureBlobToAnaplan",
                "sources": [
                    {
                        "connectionId": connection_id,
                        "type": "AzureBlob",
                        "file": "dummy/liquor_sales.csv",
                    }
                ],
                "targets": [
                    {"type": "Anaplan", "actionId": "112000000064", "fileId": "113000000055"}
                ],
            },
            {
                "type": "AzureBlobToAnaplan",
                "sources": [
                    {
                        "connectionId": connection_id,
                        "type": "AzureBlob",
                        "file": "dummy/liquor_sales.csv",
                    }
                ],
                "targets": [
                    {"type": "Anaplan", "actionId": "112000000064", "fileId": "113000000055"}
                ],
            },
        ],
    }


@pytest.fixture
def process_integration_dict(name):
    return {
        "name": name,
        "version": "2.0",
        "processId": 118000000012,
        "workspaceId": getenv("ANAPLAN_SDK_TEST_WORKSPACE_ID"),
        "modelId": getenv("ANAPLAN_SDK_TEST_MODEL_ID"),
    }


@pytest.fixture
def schedule_pydantic(name):
    return ScheduleInput(
        name=name,
        timezone="Europe/Paris",
        start_date="2027-01-01",
        end_date="2029-01-01",
        days_of_week=[0, 2, 5],
        type="daily",
        time="12:00",
    )


@pytest.fixture
def schedule_dict(name):
    return {
        "name": name,
        "type": "daily",
        "timezone": "Europe/Paris",
        "time": "12:00",
        "daysOfWeek": [0, 2, 5],
        "startDate": "2027-01-01",
        "endDate": "2029-01-01",
    }


@pytest.fixture
def notification_pydantic(name):
    return NotificationInput(
        integration_ids=[],
        channels=["in_app"],
        notifications=NotificationConfigInput(
            config=[
                NotificationItemInput(
                    type="full_failure", users=["8a868cd97f8f98a3017fe45cbdc65e25"]
                )
            ]
        ),
    )


@pytest.fixture
def notification_dict(name):
    return {
        "integrationIds": [],
        "channels": ["in_app"],
        "notifications": {
            "config": [{"type": "full_failure", "users": ["8a868cd97f8f98a3017fe45cbdc65e25"]}]
        },
    }


@pytest.fixture
def flow_pydantic(name):
    return FlowInput(
        name=name,
        steps=[
            FlowStepInput(referrer="840ccd8a279a454d99577d9538f24f09"),
            FlowStepInput(
                referrer="c0fa795faac047468a59c8dbe3752d75",
                depends_on=["840ccd8a279a454d99577d9538f24f09"],
            ),
        ],
    )


@pytest.fixture
def flow_dict(name):
    return {
        "name": name,
        "steps": [
            {"referrer": "840ccd8a279a454d99577d9538f24f09"},
            {
                "referrer": "c0fa795faac047468a59c8dbe3752d75",
                "dependsOn": ["840ccd8a279a454d99577d9538f24f09"],
            },
        ],
    }



import logging
from typing import Callable

import httpx

from .exceptions import AnaplanException, InvalidCredentialsException

logger = logging.getLogger("anaplan_sdk")


class _BaseOauth:
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        redirect_uri: str,
        authorization_url: str = "https://us1a.app.anaplan.com/auth/prelogin",
        token_url: str = "https://us1a.app.anaplan.com/oauth/token",
        validation_url: str = "https://auth.anaplan.com/token/validate",
        scope: str = "openid profile email offline_access",
        state_generator: Callable[[], str] | None = None,
    ):
        """
        Initializes the OAuth Client. This class provides the two utilities needed to implement
        the OAuth 2.0 authorization code flow for user-facing Web Applications. It differs from the
        other Authentication Strategies in this SDK in two main ways:

        1. You must implement the actual authentication flow in your application. You cannot pass
        the credentials directly to the `Client` or `AsyncClient`, and this class does not
        implement the SDK internal authentication flow, i.e. it does not subclass `httpx.Auth`.

        2. You then simply pass the resulting token to the `Client` or `AsyncClient`, rather than
        passing the credentials directly, which will internally construct an `httpx.Auth` instance

        Note that this class exist for convenience only, and you can implement the OAuth 2.0 Flow
        yourself in your preferred library, or bring an existing implementation. For details on the
        Anaplan OAuth 2.0 Flow, see the [the Docs](https://anaplanoauth2service.docs.apiary.io/#reference/overview-of-the-authorization-code-grant).
        :param client_id: The client ID of your Anaplan Oauth 2.0 application. This Application
               must be an Authorization Code Grant application.
        :param client_secret: The client secret of your Anaplan Oauth 2.0 application.
        :param redirect_uri: The URL to which the user will be redirected after authorizing the
               application.
        :param authorization_url: The URL to which the user will be redirected to authorize the
               application. Defaults to the Anaplan Prelogin Page, where the user can select the
               login method.
        :param token_url: The URL to post the authorization code to in order to fetch the access
               token.
        :param validation_url: The URL to validate the access token.
        :param scope: The scope of the access request.
        :param state_generator: A callable that generates a random state string. You can optionally
                pass this if you need to customize the state generation logic. If not provided,
                the state will be generated by `oauthlib`.
        """
        self._client_id = client_id
        self._client_secret = client_secret
        self._redirect_uri = redirect_uri
        self._authorization_url = authorization_url
        self._token_url = token_url
        self._validation_url = validation_url
        self._scope = scope

        try:
            from oauthlib.oauth2 import WebApplicationClient
        except ImportError as e:
            raise AnaplanException(
                "oauthlib is not available. Please install anaplan-sdk with the oauth extra "
                "`pip install anaplan-sdk[oauth]` or install oauthlib separately."
            ) from e
        self._oauth = WebApplicationClient(client_id=client_id, client_secret=client_secret)
        self._state_generator = state_generator if state_generator else self._oauth.state_generator

    def authorization_url(
        self, authorization_url: str | None = None, state: str | None = None
    ) -> tuple[str, str]:
        """
        Generates the authorization URL for the OAuth 2.0 flow.
        :param authorization_url: You can optionally pass a custom authorization URL. This is
               useful if you want to redirect i.e. redirect the user directly to the Anaplan login
               page rather than the Prelogin page in only one scenario, while still reusing the
               Client.
        :param state: You can optionally pass a custom state string. If not provided, a random
                state string will be generated by the `oauthlib` library, or by the
                `state_generator` callable if provided.
        :return: A tuple containing the authorization URL and the state string.
        """
        auth_url = authorization_url or self._authorization_url
        state = state or self._state_generator()
        url, _, _ = self._oauth.prepare_authorization_request(
            auth_url, state, self._redirect_uri, self._scope
        )
        return url, state

    def _token_request(self, authorization_response: str) -> httpx.Request:
        url, headers, body = self._oauth.prepare_token_request(
            authorization_response=authorization_response,
            token_url=self._token_url,
            redirect_url=self._redirect_uri,
            client_secret=self._client_secret,
        )
        return httpx.Request(method="POST", url=url, headers=headers, content=body)

    def _refresh_token_request(self, refresh_token: str) -> httpx.Request:
        url, headers, body = self._oauth.prepare_refresh_token_request(
            self._token_url,
            refresh_token=refresh_token,
            client_id=self._client_id,
            client_secret=self._client_secret,
        )
        return httpx.Request(method="POST", url=url, headers=headers, content=body)

    def _parse_response(self, response: httpx.Response) -> dict[str, str]:
        if response.status_code == 401:
            raise InvalidCredentialsException
        if not response.is_success:
            raise AnaplanException(
                f"Token request for Client {self._client_id} failed: "
                f"{response.status_code} {response.text}"
            )
        return response.json()


class _OAuthRequestFactory(_BaseOauth):
    def token_request(self, authorization_response: str) -> httpx.Request:
        return self._token_request(authorization_response)

    def refresh_token_request(self, refresh_token: str) -> httpx.Request:
        return self._refresh_token_request(refresh_token)


class AsyncOauth(_BaseOauth):
    """
    Asynchronous Variant of the Anaplan OAuth client for interactive OAuth Flows in Web
    Applications.
    """

    async def fetch_token(self, authorization_response: str) -> dict[str, str | int]:
        """
        Fetches the token using the authorization response from the OAuth 2.0 flow.
        :param authorization_response: The full URL that the user was redirected to after
               authorizing the application. This URL will contain the authorization code and state.
        :return: The token as a dictionary containing the access token, refresh token, scope,
                 expires_in, and type.
        """
        from oauthlib.oauth2 import OAuth2Error

        try:
            async with httpx.AsyncClient() as client:
                response = await client.send(self._token_request(authorization_response))
            return self._parse_response(response)
        except (httpx.HTTPError, ValueError, TypeError, OAuth2Error) as error:
            logger.error(error)
            raise AnaplanException("Error during token creation.") from error

    async def validate_token(self, token: str) -> dict[str, str | dict[str, str | int]]:
        """
        Validates the provided token by checking its validity with the Anaplan Authentication API.
        If the token is not valid, an `InvalidCredentialsException` is raised.
        :param token: The access token to validate.
        :return: The Token information as a dictionary containing the token's details.
        """
        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(
                    url=self._validation_url, headers={"Authorization": f"AnaplanAuthToken {token}"}
                )
            return self._parse_response(response)
        except httpx.HTTPError as error:
            logger.error(error)
            raise AnaplanException("Error during token validation.") from error

    async def refresh_token(self, refresh_token: str) -> dict[str, str | int]:
        """
        Refreshes the token using a refresh token.
        :param refresh_token: The refresh token to use for refreshing the access token.
        :return: The new token as a dictionary containing the access token, refresh token, scope,
                 expires_in, and type.
        """
        from oauthlib.oauth2 import OAuth2Error

        try:
            async with httpx.AsyncClient() as client:
                response = await client.send(self._refresh_token_request(refresh_token))
            return self._parse_response(response)
        except (httpx.HTTPError, ValueError, TypeError, OAuth2Error) as error:
            logger.error(error)
            raise AnaplanException("Error during token refresh.") from error


class Oauth(_BaseOauth):
    """
    Synchronous Variant of the Anaplan OAuth client for interactive OAuth Flows in Web
    Applications.
    """

    def fetch_token(self, authorization_response: str) -> dict[str, str | int]:
        """
        Fetches the token using the authorization response from the OAuth 2.0 flow.
        :param authorization_response: The full URL that the user was redirected to after
               authorizing the application. This URL will contain the authorization code and state.
        :return: The token as a dictionary containing the access token, refresh token, scope,
                 expires_in, and type.
        """
        from oauthlib.oauth2 import OAuth2Error

        try:
            url, headers, body = self._oauth.prepare_token_request(
                authorization_response=authorization_response,
                token_url=self._token_url,
                redirect_url=self._redirect_uri,
                client_secret=self._client_secret,
            )
            with httpx.Client() as client:
                response = client.post(url=url, headers=headers, content=body)
            return self._parse_response(response)
        except (httpx.HTTPError, ValueError, TypeError, OAuth2Error) as error:
            logger.error(error)
            raise AnaplanException("Error during token creation.") from error

    def validate_token(self, token: str) -> dict[str, str | dict[str, str | int]]:
        """
        Validates the provided token by checking its validity with the Anaplan Authentication API.
        If the token is not valid, an `InvalidCredentialsException` is raised.
        :param token: The access token to validate.
        :return: The Token information as a dictionary containing the token's details.
        """
        try:
            with httpx.Client() as client:
                response = client.get(
                    url=self._validation_url, headers={"Authorization": f"AnaplanAuthToken {token}"}
                )
            return self._parse_response(response)
        except httpx.HTTPError as error:
            logger.error(error)
            raise AnaplanException("Error during token validation.") from error

    def refresh_token(self, refresh_token: str) -> dict[str, str | int]:
        """
        Refreshes the token using a refresh token.
        :param refresh_token: The refresh token to use for refreshing the access token.
        :return: The new token as a dictionary containing the access token, refresh token, scope,
                 expires_in, and type.
        """
        from oauthlib.oauth2 import OAuth2Error

        try:
            url, headers, body = self._oauth.prepare_refresh_token_request(
                self._token_url,
                refresh_token=refresh_token,
                client_id=self._client_id,
                client_secret=self._client_secret,
            )
            with httpx.Client() as client:
                response = client.post(url=url, headers=headers, content=body)
            return self._parse_response(response)
        except (httpx.HTTPError, ValueError, TypeError, OAuth2Error) as error:
            logger.error(error)
            raise AnaplanException("Error during token refresh.") from error



import logging
from typing import Any, Literal

from anaplan_sdk._services import _HttpService
from anaplan_sdk._utils import (
    connection_body_payload,
    construct_payload,
    integration_payload,
    schedule_payload,
)
from anaplan_sdk.models.cloud_works import (
    Connection,
    ConnectionBody,
    ConnectionInput,
    Integration,
    IntegrationInput,
    IntegrationProcessInput,
    NotificationConfig,
    NotificationInput,
    RunError,
    RunStatus,
    RunSummary,
    ScheduleInput,
    SingleIntegration,
)

from ._cw_flow import _FlowClient

logger = logging.getLogger("anaplan_sdk")


class _CloudWorksClient:
    def __init__(self, http: _HttpService) -> None:
        self._http = http
        self._url = "https://api.cloudworks.anaplan.com/2/0/integrations"
        self._flow = _FlowClient(self._http)

    @property
    def flows(self) -> _FlowClient:
        """
        Access the Integration Flow APIs.
        """
        return self._flow

    def get_connections(self) -> list[Connection]:
        """
        List all Connections available in CloudWorks.
        :return: A list of connections.
        """
        return [
            Connection.model_validate(e)
            for e in self._http.get_paginated(f"{self._url}/connections", "connections")
        ]

    def create_connection(self, con_info: ConnectionInput | dict[str, Any]) -> str:
        """
        Create a new connection in CloudWorks.
        :param con_info: The connection information. This can be a ConnectionInput instance or a
               dictionary as per the documentation. If a dictionary is passed, it will be validated
               against the ConnectionInput model before sending the request.
        :return: The ID of the new connection.
        """
        res = self._http.post(
            f"{self._url}/connections", json=construct_payload(ConnectionInput, con_info)
        )
        connection_id = res["connections"]["connectionId"]
        logger.info(f"Created connection '{connection_id}'.")
        return connection_id

    def update_connection(self, con_id: str, con_info: ConnectionBody | dict[str, Any]) -> None:
        """
        Update an existing connection in CloudWorks.
        :param con_id: The ID of the connection to update.
        :param con_info: The name and details of the connection. You must pass all the same details
               as when initially creating the connection again. If you want to update only some of
               the details, use the `patch_connection` method instead.
        """
        self._http.put(f"{self._url}/connections/{con_id}", json=connection_body_payload(con_info))

    def patch_connection(self, con_id: str, body: dict[str, Any]) -> None:
        """
        Update an existing connection in CloudWorks.
        :param con_id: The ID of the connection to update.
        :param body: The name and details of the connection. You can pass all the same details as
               when initially creating the connection again, or just any one of them.
        """
        self._http.patch(f"{self._url}/connections/{con_id}", json=body)

    def delete_connection(self, con_id: str) -> None:
        """
        Delete an existing connection in CloudWorks.
        :param con_id: The ID of the connection to delete.
        """
        self._http.delete(f"{self._url}/connections/{con_id}")
        logger.info(f"Deleted connection '{con_id}'.")

    def get_integrations(
        self, sort_by: Literal["name"] | None = None, descending: bool = False
    ) -> list[Integration]:
        """
        List all integrations in CloudWorks.
        :param sort_by: The field to sort the results by.
        :param descending: If True, the results will be sorted in descending order.
        :return: A list of integrations.
        """
        params = {"sortBy": f"{'-' if descending else ''}{sort_by}"} if sort_by else None
        return [
            Integration.model_validate(e)
            for e in self._http.get_paginated(f"{self._url}", "integrations", params=params)
        ]

    def get_integration(self, integration_id: str) -> SingleIntegration:
        """
        Get the details of a specific integration in CloudWorks.

        **Note: This will not include the integration type! While present when listing integrations,
        the integration type is not included in the details of a single integration.**
        :param integration_id: The ID of the integration to retrieve.
        :return: The details of the integration, without the integration type.
        """
        return SingleIntegration.model_validate(
            (self._http.get(f"{self._url}/{integration_id}"))["integration"]
        )

    def create_integration(
        self, body: IntegrationInput | IntegrationProcessInput | dict[str, Any]
    ) -> str:
        """
        Create a new integration in CloudWorks. If not specified, the integration type will be
        either "Import" or "Export" based on the source and target you provide.

        If you want to instead create a process Integration, you can do so by specifying
        the `process_id` parameter and passing several jobs. **Be careful to ensure, that all ids
        specified in the job inputs match what is defined in your model and matches the process.**
        If this is not the case, this will error, occasionally with a misleading error message,
        i.e. `XYZ is not defined in your model` even though it is, Anaplan just does not know what
        to do with it in the location you specified.

        You can also use CloudWorks Integrations to simply schedule a process. To do this, you
        can simply pass an IntegrationProcessInput instance with the process_id and no jobs. This
        will create a process integration that will run the process you specified.
        :param body: The integration information. This can be an
               IntegrationInput | IntegrationProcessInput instance or a dictionary as per the
               documentation. If a dictionary is passed, it will be validated against the
               IntegrationInput model before sending the request.
        :return: The ID of the new integration.
        """
        json = integration_payload(body)
        integration_id = (self._http.post(f"{self._url}", json=json))["integration"][
            "integrationId"
        ]
        logger.info(f"Created integration '{integration_id}'.")
        return integration_id

    def update_integration(
        self, integration_id: str, body: IntegrationInput | IntegrationProcessInput | dict[str, Any]
    ) -> None:
        """
        Update an existing integration in CloudWorks.
        :param integration_id: The ID of the integration to update.
        :param body: The name and details of the integration. You must pass all the same details
               as when initially creating the integration again. If you want to update only some
               of the details, use the `patch_integration` method instead.
        """
        json = integration_payload(body)
        self._http.put(f"{self._url}/{integration_id}", json=json)

    def run_integration(self, integration_id: str) -> str:
        """
        Run an integration in CloudWorks.
        :param integration_id: The ID of the integration to run.
        :return: The ID of the run instance.
        """
        run_id = (self._http.post_empty(f"{self._url}/{integration_id}/run"))["run"]["id"]
        logger.info(f"Started integration run '{run_id}' for integration '{integration_id}'.")
        return run_id

    def delete_integration(self, integration_id: str) -> None:
        """
        Delete an existing integration in CloudWorks.
        :param integration_id: The ID of the integration to delete.
        """
        self._http.delete(f"{self._url}/{integration_id}")
        logger.info(f"Deleted integration '{integration_id}'.")

    def get_run_history(self, integration_id: str) -> list[RunSummary]:
        """
        Get the run history of a specific integration in CloudWorks.
        :param integration_id: The ID of the integration to retrieve the run history for.
        :return: A list of run statuses.
        """
        return [
            RunSummary.model_validate(e)
            for e in (self._http.get(f"{self._url}/runs/{integration_id}"))["history_of_runs"].get(
                "runs", []
            )
        ]

    def get_run_status(self, run_id: str) -> RunStatus:
        """
        Get the status of a specific run in CloudWorks.
        :param run_id: The ID of the run to retrieve.
        :return: The details of the run.
        """
        return RunStatus.model_validate((self._http.get(f"{self._url}/run/{run_id}"))["run"])

    def get_run_error(self, run_id: str) -> RunError | None:
        """
        Get the error details of a specific run in CloudWorks. This exposes potential underlying
        errors like the error of the invoked action, failure dumps and other details.
        :param run_id: The ID of the run to retrieve.
        :return: The details of the run error.
        """
        run = self._http.get(f"{self._url}/runerror/{run_id}")
        return RunError.model_validate(run["runs"]) if run.get("runs") else None

    def create_schedule(
        self, integration_id: str, schedule: ScheduleInput | dict[str, Any]
    ) -> None:
        """
        Schedule an integration in CloudWorks.
        :param integration_id: The ID of the integration to schedule.
        :param schedule: The schedule information. This can be a ScheduleInput instance or a
               dictionary as per the documentation. If a dictionary is passed, it will be validated
               against the ScheduleInput model before sending the request.
        """
        self._http.post(
            f"{self._url}/{integration_id}/schedule",
            json=schedule_payload(integration_id, schedule),
        )
        logger.info(f"Created schedule for integration '{integration_id}'.")

    def update_schedule(
        self, integration_id: str, schedule: ScheduleInput | dict[str, Any]
    ) -> None:
        """
        Update an integration Schedule in CloudWorks. A schedule must already exist.
        :param integration_id: The ID of the integration to schedule.
        :param schedule: The schedule information. This can be a ScheduleInput instance or a
               dictionary as per the documentation. If a dictionary is passed, it will be validated
               against the ScheduleInput model before sending the request.
        """
        self._http.put(
            f"{self._url}/{integration_id}/schedule",
            json=schedule_payload(integration_id, schedule),
        )
        logger.info(f"Updated schedule for integration '{integration_id}'.")

    def set_schedule_status(
        self, integration_id: str, status: Literal["enabled", "disabled"]
    ) -> None:
        """
        Set the status of an integration schedule in CloudWorks. A schedule must already exist.
        :param integration_id: The ID of the integration to schedule.
        :param status: The status of the schedule. This can be either "enabled" or "disabled".
        """
        self._http.post_empty(f"{self._url}/{integration_id}/schedule/status/{status}")
        logger.info(f"Set schedule status to '{status}' for integration '{integration_id}'.")

    def delete_schedule(self, integration_id: str) -> None:
        """
        Delete an integration schedule in CloudWorks. A schedule must already exist.
        :param integration_id: The ID of the integration to schedule.
        """
        self._http.delete(f"{self._url}/{integration_id}/schedule")
        logger.info(f"Deleted schedule for integration '{integration_id}'.")

    def get_notification_config(
        self, notification_id: str | None = None, integration_id: str | None = None
    ) -> NotificationConfig:
        """
        Get the notification configuration, either by its Id, or the notification configuration
        for a specific integration. If the integration_id is specified, the notification_id
        will be ignored.
        :param notification_id: The ID of the notification configuration to retrieve.
        :param integration_id: The ID of the integration to retrieve the notification
               configuration for.
        :return: The details of the notification configuration.
        """
        if not (notification_id or integration_id):
            raise ValueError("Either notification_id or integration_id must be specified.")
        if integration_id:
            notification_id = (self.get_integration(integration_id)).notification_id
        return NotificationConfig.model_validate(
            (self._http.get(f"{self._url}/notification/{notification_id}"))["notifications"]
        )

    def create_notification_config(self, config: NotificationInput | dict[str, Any]) -> str:
        """
        Create a notification configuration for an integration in CloudWorks. This will error if
        there is already a notification configuration for the integration, which is also the case
        by default. In this case, you will want to use the `update_notification_config` method
        instead, to partially update the existing configuration or overwrite it.
        :param config: The notification configuration. This can be a NotificationInput instance or
               a dictionary as per the documentation. If a dictionary is passed, it will be
               validated against the NotificationConfig model before sending the request.
        :return: The ID of the new notification configuration.
        """
        res = self._http.post(
            f"{self._url}/notification", json=construct_payload(NotificationInput, config)
        )
        notification_id = res["notification"]["notificationId"]
        logger.info(f"Created notification configuration '{notification_id}'.")
        return notification_id

    def update_notification_config(
        self, notification_id: str, config: NotificationInput | dict[str, Any]
    ) -> None:
        """
        Update a notification configuration for an integration in CloudWorks. You cannot pass empty
        values or nulls to any of the fields If you want to for e.g. override an existing list of
        users with an empty one, you must delete the notification configuration and create a new
        one with only the values you want to keep.
        :param notification_id: The ID of the notification configuration to update.
        :param config: The notification configuration. This can be a NotificationInput instance or
               a dictionary as per the documentation. If a dictionary is passed, it will be
               validated against the NotificationConfig model before sending the request.
        """
        self._http.put(
            f"{self._url}/notification/{notification_id}",
            json=construct_payload(NotificationInput, config),
        )

    def delete_notification_config(
        self, notification_id: str | None = None, integration_id: str | None = None
    ) -> None:
        """
        Delete a notification configuration for an integration in CloudWorks, either by its Id, or
        the notification configuration for a specific integration. If the integration_id is
        specified, the notification_id will be ignored.
        :param notification_id: The ID of the notification configuration to delete.
        :param integration_id: The ID of the integration to delete the notification config of.
        """
        if not (notification_id or integration_id):
            raise ValueError("Either notification_id or integration_id must be specified.")
        if integration_id:
            notification_id = (self.get_integration(integration_id)).notification_id
        self._http.delete(f"{self._url}/notification/{notification_id}")
        logger.info(f"Deleted notification configuration '{notification_id}'.")

    def get_import_error_dump(self, run_id: str) -> bytes:
        """
        Get the error dump of a specific import run in CloudWorks. Calling this for a run_id that
        did not generate any failure dumps will produce an error.

        **Note that if you need the error dump of an action in a process, you must use the
        `get_process_error_dump` method instead.**
        :param run_id: The ID of the run to retrieve.
        :return: The error dump.
        """
        return self._http.get_binary(f"{self._url}/run/{run_id}/dump")

    def get_process_error_dump(self, run_id: str, action_id: int | str) -> bytes:
        """
        Get the error dump of a specific import run in CloudWorks, that is part of a process.
        Calling this for a run_id that did not generate any failure dumps will produce an error.
        :param run_id: The ID of the run to retrieve.
        :param action_id: The ID of the action to retrieve. This can be found in the RunError.
        :return: The error dump.
        """
        return self._http.get_binary(f"{self._url}/run/{run_id}/process/import/{action_id}/dumps")



from anaplan_sdk.models import User


def test_list_users(client):
    users, search = client.audit.get_users(), client.audit.get_users("vinzenz")
    assert isinstance(users, list)
    assert all(isinstance(user, User) for user in users)
    assert all(isinstance(user, User) for user in search)
    assert len(users) > 0
    assert len(search) > 0
    assert len(search) < len(users)


def test_get_user(client):
    user = client.audit.get_user()
    assert isinstance(user, User)


def test_events(client):
    events = client.audit.get_events(1)
    assert isinstance(events, list)



from datetime import datetime
from uuid import uuid4

from anaplan_sdk import Client
from anaplan_sdk.models import (
    ModelRevision,
    ReportTask,
    Revision,
    SummaryReport,
    SyncTask,
    TaskSummary,
)


def test_change_model_status(alm_client: Client):
    alm_client.alm.change_model_status("offline")


def test_list_revisions(alm_client: Client):
    revisions = alm_client.alm.get_revisions()
    assert isinstance(revisions, list)
    assert all(isinstance(rev, Revision) for rev in revisions)


def test_get_latest_revision(alm_client: Client):
    rev = alm_client.alm.get_latest_revision()
    assert isinstance(rev, Revision)


def test_list_syncable_revisions(alm_client: Client, alm_src_model_id: str):
    revisions = alm_client.alm.get_syncable_revisions(alm_src_model_id)
    assert isinstance(revisions, list)
    assert all(isinstance(rev, Revision) for rev in revisions)


def test_create_revision(alm_src_client: Client):
    alm_src_client.tr.insert_list_items(
        101000000005, [{"name": str(uuid4()), "code": str(uuid4())}]
    )
    name, desc = datetime.now().strftime("%Y-%m-%d %H:%M:%S"), "Test Sync Revision."
    revision = alm_src_client.alm.create_revision(name, desc)
    assert isinstance(revision, Revision)
    assert revision.name == name
    assert revision.description == desc


def test_create_comparison_report(
    alm_client: Client, alm_src_client: Client, alm_src_model_id: str
):
    src_rev, latest_rev = (
        alm_src_client.alm.get_latest_revision(),
        alm_client.alm.get_latest_revision(),
    )
    report_task = alm_client.alm.create_comparison_report(
        src_rev.id, alm_src_model_id, latest_rev.id
    )
    assert isinstance(report_task, ReportTask)
    report = alm_client.alm.get_comparison_report(report_task)
    assert report is not None


def test_create_comparison_summary_task(
    alm_client: Client, alm_src_client: Client, alm_src_model_id: str
):
    src_rev, latest_rev = (
        alm_src_client.alm.get_latest_revision(),
        alm_client.alm.get_latest_revision(),
    )
    report_task = alm_client.alm.create_comparison_summary(
        src_rev.id, alm_src_model_id, latest_rev.id, False
    )
    assert isinstance(report_task, ReportTask)


def test_create_comparison_summary(
    alm_client: Client, alm_src_client: Client, alm_src_model_id: str
):
    src_rev, latest_rev = (
        alm_src_client.alm.get_latest_revision(),
        alm_client.alm.get_latest_revision(),
    )
    report = alm_client.alm.create_comparison_summary(src_rev.id, alm_src_model_id, latest_rev.id)
    assert isinstance(report, SummaryReport)
    assert sum((report.totals.created, report.totals.modified, report.totals.deleted)) > 0


def test_sync_models(alm_client: Client, alm_src_client: Client, alm_src_model_id: str):
    src_rev, latest_rev = (
        alm_src_client.alm.get_latest_revision(),
        alm_client.alm.get_latest_revision(),
    )
    sync_task = alm_client.alm.sync_models(src_rev.id, alm_src_model_id, latest_rev.id)
    assert isinstance(sync_task, SyncTask)
    assert sync_task.result.source_revision_id == src_rev.id
    assert sync_task.result.target_revision_id == latest_rev.id


def test_list_models_for_revision(alm_client: Client):
    rev = alm_client.alm.get_latest_revision()
    model_revs = alm_client.alm.get_models_for_revision(rev.id)
    assert isinstance(model_revs, list)
    assert all(isinstance(rev, ModelRevision) for rev in model_revs)


def test_list_sync_tasks(alm_client: Client):
    tasks = alm_client.alm.get_sync_tasks()
    assert isinstance(tasks, list)
    assert all(isinstance(task, TaskSummary) for task in tasks)



import logging
import multiprocessing
from concurrent.futures import ThreadPoolExecutor
from copy import copy
from typing import Iterator, Literal

import httpx
from typing_extensions import Self

from anaplan_sdk._auth import _create_auth
from anaplan_sdk._services import _HttpService
from anaplan_sdk._utils import action_url, models_url, sort_params
from anaplan_sdk.exceptions import AnaplanActionError, InvalidIdentifierException
from anaplan_sdk.models import (
    Action,
    Export,
    File,
    Import,
    Model,
    ModelDeletionResult,
    Process,
    TaskStatus,
    TaskSummary,
    Workspace,
)

from ._alm import _AlmClient
from ._audit import _AuditClient
from ._cloud_works import _CloudWorksClient
from ._scim import _ScimClient
from ._transactional import _TransactionalClient

SortBy = Literal["id", "name"] | None

logger = logging.getLogger("anaplan_sdk")


class Client:
    """
    Synchronous Anaplan Client. For guides and examples
    refer to https://vinzenzklass.github.io/anaplan-sdk.
    """

    def __init__(
        self,
        workspace_id: str | None = None,
        model_id: str | None = None,
        user_email: str | None = None,
        password: str | None = None,
        certificate: str | bytes | None = None,
        private_key: str | bytes | None = None,
        private_key_password: str | bytes | None = None,
        token: str | None = None,
        auth: httpx.Auth | None = None,
        timeout: float | httpx.Timeout = 30,
        retry_count: int = 2,
        backoff: float = 1.0,
        backoff_factor: float = 2.0,
        page_size: int = 5_000,
        status_poll_delay: int = 1,
        upload_parallel: bool = True,
        upload_chunk_size: int = 25_000_000,
        allow_file_creation: bool = False,
        **httpx_kwargs,
    ) -> None:
        """
        Synchronous Anaplan Client. For guides and examples
        refer to https://vinzenzklass.github.io/anaplan-sdk.

        :param workspace_id: The Anaplan workspace Id. You can copy this from the browser URL or
               find them using an HTTP Client like Postman, Paw, Insomnia etc.
        :param model_id: The identifier of the model.
        :param user_email: A valid email registered with the Anaplan Workspace you are attempting
               to access.
        :param password: Password for the given `user_email` for basic Authentication.
        :param certificate: The certificate content or the absolute path to the certificate file.
        :param private_key: The private key content or the absolute path to the private key file.
        :param private_key_password: The password to access the private key file. This is only
               considered if you provided a private key file and it password-protected.
        :param token: An Anaplan API Token. This will be used to authenticate the client. If
               sufficient other authentication parameters are provided, the token will be used
               until it expires, after which a new one will be created. If you provide only this
               parameter, the client will raise an error upon first authentication failure. For
               short-lived instances, such as in web applications where user specific clients are
               created, this is the recommended way to authenticate, since this has the least
               overhead.
        :param auth: You can provide a subclass of `httpx.Auth` to use for authentication. You can
               provide an instance of one of the classes provided by the SDK, or an instance of
               your own subclass of `httpx.Auth`. This will give you full control over the
               authentication process, but you will need to implement the entire authentication
               logic yourself.
        :param timeout: The timeout in seconds for the HTTP requests. Alternatively, you can pass
               an instance of `httpx.Timeout` to set the timeout for the HTTP requests.
        :param retry_count: The number of times to retry an HTTP request if it fails. Set this to 0
               to never retry. Defaults to 2, meaning each HTTP Operation will be tried a total
        :param backoff: The initial backoff time in seconds for the retry mechanism. This is the
               time to wait before the first retry.
        :param backoff_factor: The factor by which the backoff time is multiplied after each retry.
               For example, if the initial backoff is 1 second and the factor is 2, the second
               retry will wait 2 seconds, the third retry will wait 4 seconds, and so on.
               number of 2 times.
        :param page_size: The number of items to return per page when paginating through results.
               Defaults to 5000. This is the maximum number of items that can be returned per
               request. If you pass a value greater than 5000, it will be capped to 5000.
        :param status_poll_delay: The delay between polling the status of a task.
        :param upload_parallel: Whether to upload chunks in parallel when uploading files.
        :param upload_chunk_size: The size of the chunks to upload. This is the maximum size of
               each chunk. Defaults to 25MB.
        :param allow_file_creation: Whether to allow the creation of new files. Defaults to False
               since this is typically unintentional and may well be unwanted behaviour in the API
               altogether. A file that is created this way will not be referenced by any action in
               anaplan until manually assigned so there is typically no value in dynamically
               creating new files and uploading content to them.
        :param httpx_kwargs: Additional keyword arguments to pass to the `httpx.Client`.
               This can be used to set additional options such as proxies, headers, etc. See
               https://www.python-httpx.org/api/#client for the full list of arguments.
        """
        auth = auth or _create_auth(
            token=token,
            user_email=user_email,
            password=password,
            certificate=certificate,
            private_key=private_key,
            private_key_password=private_key_password,
        )
        _client = httpx.Client(auth=auth, timeout=timeout, **httpx_kwargs)
        self._http = _HttpService(
            _client,
            retry_count=retry_count,
            backoff=backoff,
            backoff_factor=backoff_factor,
            page_size=page_size,
            poll_delay=status_poll_delay,
        )
        self._retry_count = retry_count
        self._workspace_id = workspace_id
        self._model_id = model_id
        self._url = f"https://api.anaplan.com/2/0/workspaces/{workspace_id}/models/{model_id}"
        self._transactional_client = (
            _TransactionalClient(self._http, model_id) if model_id else None
        )
        self._alm_client = _AlmClient(self._http, model_id) if model_id else None
        self._audit_client = _AuditClient(self._http)
        self._scim_client = _ScimClient(self._http)
        self._cloud_works = _CloudWorksClient(self._http)
        self._thread_count = multiprocessing.cpu_count()
        self.status_poll_delay = status_poll_delay
        self.upload_parallel = upload_parallel
        self.upload_chunk_size = upload_chunk_size
        self.allow_file_creation = allow_file_creation
        logger.debug(f"Initialized Client with workspace_id={workspace_id}, model_id={model_id}")

    def with_model(self, model_id: str | None = None, workspace_id: str | None = None) -> Self:
        """
        Create a new instance of the Client with the given model and workspace Ids. **This creates
        a copy of the current client. The current instance remains unchanged.**
        :param workspace_id: The workspace Id to use or None to use the existing workspace Id.
        :param model_id: The model Id to use or None to use the existing model Id.
        :return: A new instance of the Client.
        """
        client = copy(self)
        client._workspace_id = workspace_id or self._workspace_id
        client._model_id = model_id or self._model_id
        logger.debug(
            f"Creating a new Client from existing instance "
            f"with workspace_id={client._workspace_id}, model_id={client._model_id}."
        )
        client._url = (
            "https://api.anaplan.com/2/0/workspaces"
            f"/{client._workspace_id}/models/{client._model_id}"
        )
        client._transactional_client = _TransactionalClient(self._http, client._model_id)
        client._alm_client = _AlmClient(self._http, client._model_id)
        return client

    @property
    def audit(self) -> _AuditClient:
        """
        The Audit Client provides access to the Anaplan Audit API.
        For details, see https://vinzenzklass.github.io/anaplan-sdk/guides/audit/.
        """
        return self._audit_client

    @property
    def cw(self) -> _CloudWorksClient:
        """
        The Cloud Works Client provides access to the Anaplan Cloud Works API.
        For details, see https://vinzenzklass.github.io/anaplan-sdk/guides/cloud_works/.
        """
        return self._cloud_works

    @property
    def tr(self) -> _TransactionalClient:
        """
        The Transactional Client provides access to the Anaplan Transactional API. This is useful
        for more advanced use cases where you need to interact with the Anaplan Model in a more
        granular way.

        If you instantiated the client without the field `model_id`, this will raise a
        `ValueError`, since none of the endpoints can be invoked without the model Id.
        :return: The Transactional Client.
        """
        if not self._transactional_client:
            raise ValueError(
                "Cannot use the Transactional Client (Anaplan Transactional API) "
                "without field `model_id`. Make sure the instance you are trying to call this on "
                "is instantiated correctly with a valid `model_id`."
            )
        return self._transactional_client

    @property
    def alm(self) -> _AlmClient:
        """
        **To use the Application Lifecycle Management (ALM) API, you need a Professional or
        Enterprise subscription.**

        The ALM Client provides access to the Anaplan ALM API. This is useful for more advanced use
        cases where you need retrieve Meta Information for yours models, read or create revisions,
        spawn sync tasks or generate comparison reports.

        :return: The ALM Client.
        """
        if not self._alm_client:
            raise ValueError(
                "Cannot use the ALM Client (Anaplan ALM API) "
                "without field `model_id`. Make sure the instance you are trying to call this on "
                "is instantiated correctly with a valid `model_id`."
            )
        return self._alm_client

    @property
    def scim(self) -> _ScimClient:
        """
        To use the SCIM API, you must be User Admin. The SCIM API allows managing internal users.
        Visiting users are excluded from the SCIM API.
        :return: The SCIM Client.
        """
        return self._scim_client

    def get_workspaces(
        self,
        search_pattern: str | None = None,
        sort_by: Literal["size_allowance", "name"] | None = None,
        descending: bool = False,
    ) -> list[Workspace]:
        """
        Lists all the Workspaces the authenticated user has access to.
        :param search_pattern:  **Caution: This is an undocumented Feature and may behave
               unpredictably. It requires the Tenant Admin role. For non-admin users, it is
               ignored.** Optionally filter for specific workspaces. When provided,
               case-insensitive matches workspaces with names containing this string.
               You can use the wildcards `%` for 0-n characters, and `_` for exactly 1 character.
               When None (default), returns all users.
        :param sort_by: The field to sort the results by.
        :param descending: If True, the results will be sorted in descending order.
        :return: The List of Workspaces.
        """
        params = {"tenantDetails": "true"} | sort_params(sort_by, descending)
        if search_pattern:
            params["s"] = search_pattern
        res = self._http.get_paginated(
            "https://api.anaplan.com/2/0/workspaces", "workspaces", params=params
        )
        return [Workspace.model_validate(e) for e in res]

    def get_models(
        self,
        only_in_workspace: bool | str = False,
        search_pattern: str | None = None,
        sort_by: Literal["active_state", "name"] | None = None,
        descending: bool = False,
    ) -> list[Model]:
        """
        Lists all the Models the authenticated user has access to.
        :param only_in_workspace: If True, only lists models in the workspace provided when
               instantiating the client. If a string is provided, only lists models in the workspace
               with the given Id. If False (default), lists models in all workspaces the user
        :param search_pattern:  **Caution: This is an undocumented Feature and may behave
               unpredictably. It requires the Tenant Admin role. For non-admin users, it is
               ignored.** Optionally filter for specific models. When provided,
               case-insensitive matches model names containing this string.
               You can use the wildcards `%` for 0-n characters, and `_` for exactly 1 character.
               When None (default), returns all models.
        :param sort_by: The field to sort the results by.
        :param descending: If True, the results will be sorted in descending order.
        :return: The List of Models.
        """
        params = {"modelDetails": "true"} | sort_params(sort_by, descending)
        if search_pattern:
            params["s"] = search_pattern
        res = self._http.get_paginated(
            models_url(only_in_workspace, self._workspace_id), "models", params=params
        )
        return [Model.model_validate(e) for e in res]

    def delete_models(self, model_ids: list[str]) -> ModelDeletionResult:
        """
        Delete the given Models. Models need to be closed before they can be deleted. If one of the
        deletions fails, the other deletions will still be attempted and may complete.
        :param model_ids: The list of Model identifiers to delete.
        :return:
        """
        logger.info(f"Deleting Models: {', '.join(model_ids)}.")
        res = self._http.post(
            f"https://api.anaplan.com/2/0/workspaces/{self._workspace_id}/bulkDeleteModels",
            json={"modelIdsToDelete": model_ids},
        )
        return ModelDeletionResult.model_validate(res)

    def get_files(
        self, sort_by: Literal["id", "name"] = "id", descending: bool = False
    ) -> list[File]:
        """
        Lists all the Files in the Model.
        :param sort_by: The field to sort the results by.
        :param descending: If True, the results will be sorted in descending order.
        :return: The List of Files.
        """
        res = self._http.get_paginated(
            f"{self._url}/files", "files", params=sort_params(sort_by, descending)
        )
        return [File.model_validate(e) for e in res]

    def get_actions(self, sort_by: SortBy = None, descending: bool = False) -> list[Action]:
        """
        Lists all the Actions in the Model. This will only return the Actions listed under
        `Other Actions` in Anaplan. For Imports, exports, and processes, see their respective
        methods instead.
        :param sort_by: The field to sort the results by.
        :param descending: If True, the results will be sorted in descending order.
        :return: The List of Actions.
        """
        res = self._http.get_paginated(
            f"{self._url}/actions", "actions", params=sort_params(sort_by, descending)
        )
        return [Action.model_validate(e) for e in res]

    def get_processes(self, sort_by: SortBy = None, descending: bool = False) -> list[Process]:
        """
        Lists all the Processes in the Model.
        :param sort_by: The field to sort the results by.
        :param descending: If True, the results will be sorted in descending order.
        :return: The List of Processes.
        """
        res = self._http.get_paginated(
            f"{self._url}/processes", "processes", params=sort_params(sort_by, descending)
        )
        return [Process.model_validate(e) for e in res]

    def get_imports(self, sort_by: SortBy = None, descending: bool = False) -> list[Import]:
        """
        Lists all the Imports in the Model.
        :param sort_by: The field to sort the results by.
        :param descending: If True, the results will be sorted in descending order.
        :return: The List of Imports.
        """
        res = self._http.get_paginated(
            f"{self._url}/imports", "imports", params=sort_params(sort_by, descending)
        )
        return [Import.model_validate(e) for e in res]

    def get_exports(self, sort_by: SortBy = None, descending: bool = False) -> list[Export]:
        """
        Lists all the Exports in the Model.
        :param sort_by: The field to sort the results by.
        :param descending: If True, the results will be sorted in descending order.
        :return: The List of Exports.
        """
        res = self._http.get_paginated(
            f"{self._url}/exports", "exports", params=sort_params(sort_by, descending)
        )
        return [Export.model_validate(e) for e in res]

    def run_action(self, action_id: int, wait_for_completion: bool = True) -> TaskStatus:
        """
        Runs the Action and validates the spawned task. If the Action fails or completes with
        errors, this will raise an AnaplanActionError. Failed Tasks are often not something you
        can recover from at runtime and often require manual changes in Anaplan, i.e. updating the
        mapping of an Import or similar.
        :param action_id: The identifier of the Action to run. Can be any Anaplan Invokable;
               Processes, Imports, Exports, Other Actions.
        :param wait_for_completion: If True, the method will poll the task status and not return
               until the task is complete. If False, it will spawn the task and return immediately.
        """
        body = {"localeName": "en_US"}
        res = self._http.post(f"{self._url}/{action_url(action_id)}/{action_id}/tasks", json=body)
        task_id = res["task"]["taskId"]
        logger.info(f"Invoked Action '{action_id}', spawned Task: '{task_id}'.")

        if not wait_for_completion:
            return TaskStatus.model_validate(self.get_task_status(action_id, task_id))
        status = self._http.poll_task(self.get_task_status, action_id, task_id)
        if status.task_state == "COMPLETE" and not status.result.successful:
            logger.error(f"Task '{task_id}' completed with errors.")
            raise AnaplanActionError(f"Task '{task_id}' completed with errors.")

        logger.info(f"Task '{task_id}' of Action '{action_id}' completed successfully.")
        return status

    def get_file(self, file_id: int) -> bytes:
        """
        Retrieves the content of the specified file.
        :param file_id: The identifier of the file to retrieve.
        :return: The content of the file.
        """
        chunk_count = self._file_pre_check(file_id)
        logger.info(f"File {file_id} has {chunk_count} chunks.")
        if chunk_count <= 1:
            return self._http.get_binary(f"{self._url}/files/{file_id}")
        with ThreadPoolExecutor(max_workers=self._thread_count) as executor:
            chunks = executor.map(
                self._http.get_binary,
                (f"{self._url}/files/{file_id}/chunks/{i}" for i in range(chunk_count)),
            )
            return b"".join(chunks)

    def get_file_stream(self, file_id: int, batch_size: int = 1) -> Iterator[bytes]:
        """
        Retrieves the content of the specified file as a stream of chunks. The chunks are yielded
        one by one, so you can process them as they arrive. This is useful for large files where
        you don't want to or cannot load the entire file into memory at once.
        :param file_id: The identifier of the file to retrieve.
        :param batch_size: Number of chunks to fetch concurrently. If > 1, n chunks will be fetched
               concurrently. This still yields each chunk individually, only the requests are
               batched. If 1 (default), each chunk is fetched sequentially.
        :return: A generator yielding the chunks of the file.
        """
        chunk_count = self._file_pre_check(file_id)
        logger.info(f"File {file_id} has {chunk_count} chunks.")
        if chunk_count <= 1:
            yield self._http.get_binary(f"{self._url}/files/{file_id}")
            return

        with ThreadPoolExecutor(max_workers=batch_size) as executor:
            for batch_start in range(0, chunk_count, batch_size):
                batch_chunks = executor.map(
                    self._http.get_binary,
                    (
                        f"{self._url}/files/{file_id}/chunks/{i}"
                        for i in range(batch_start, min(batch_start + batch_size, chunk_count))
                    ),
                )
                for chunk in batch_chunks:
                    yield chunk

    def upload_file(self, file_id: int, content: str | bytes) -> None:
        """
        Uploads the content to the specified file. If there are several chunks, upload of
        individual chunks are uploaded concurrently.

        :param file_id: The identifier of the file to upload to.
        :param content: The content to upload. **This Content will be compressed before uploading.
               If you are passing the Input as bytes, pass it uncompressed.**
        """
        if isinstance(content, str):
            content = content.encode()
        chunks = [
            content[i : i + self.upload_chunk_size]
            for i in range(0, len(content), self.upload_chunk_size)
        ]
        logger.info(f"Content for file '{file_id}' will be uploaded in {len(chunks)} chunks.")
        self._set_chunk_count(file_id, len(chunks))
        if self.upload_parallel:
            with ThreadPoolExecutor(max_workers=self._thread_count) as executor:
                executor.map(
                    self._upload_chunk, (file_id,) * len(chunks), range(len(chunks)), chunks
                )
        else:
            for index, chunk in enumerate(chunks):
                self._upload_chunk(file_id, index, chunk)
        logger.info(f"Completed upload for file '{file_id}'.")

    def upload_file_stream(
        self, file_id: int, content: Iterator[str | bytes], batch_size: int = 1
    ) -> None:
        """
        Uploads the content to the specified file as a stream of chunks. This is useful either for
        large files where you don't want to or cannot load the entire file into memory at once, or
        if you simply do not know the number of chunks ahead of time and instead just want to pass
        on chunks i.e. consumed from a queue until it is exhausted. In this case, you can pass a
        generator that yields the chunks of the file one by one to this method.

        :param file_id: The identifier of the file to upload to.
        :param content: An Iterator or AsyncIterator yielding the chunks of the file. You can pass
               any Iterator, but you will most likely want to pass a Generator.
        :param batch_size: Number of chunks to upload concurrently. If > 1, n chunks will be
               uploaded concurrently. This can be useful if you either do not control the chunk
               size, or if you want to keep the chunk size small but still want some concurrency.
        """
        logger.info(f"Starting upload stream for file '{file_id}' with batch size {batch_size}.")
        self._set_chunk_count(file_id, -1)
        indices, chunks = [], []
        with ThreadPoolExecutor(max_workers=batch_size) as executor:
            for index, chunk in enumerate(content):
                indices.append(index)
                chunks.append(chunk)
                if len(indices) == max(batch_size, 1):
                    list(
                        executor.map(self._upload_chunk, (file_id,) * len(indices), indices, chunks)
                    )
                    logger.info(
                        f"Completed upload stream batch of size {batch_size} for file {file_id}."
                    )
                    indices, chunks = [], []

            if indices:
                executor.map(self._upload_chunk, (file_id,) * len(indices), indices, chunks)
        logger.info(
            f"Completed final upload stream batch of size {len(indices)} for file {file_id}."
        )
        self._http.post(f"{self._url}/files/{file_id}/complete", json={"id": file_id})
        logger.info(f"Completed upload stream for '{file_id}'.")

    def upload_and_import(
        self, file_id: int, content: str | bytes, action_id: int, wait_for_completion: bool = True
    ) -> TaskStatus:
        """
        Convenience wrapper around `upload_file()` and `run_action()` to upload content to a file
        and run an import action in one call.
        :param file_id: The identifier of the file to upload to.
        :param content: The content to upload. **This Content will be compressed before uploading.
               If you are passing the Input as bytes, pass it uncompressed to avoid redundant
               work.**
        :param action_id: The identifier of the action to run after uploading the content.
        :param wait_for_completion: If True, the method will poll the import task status and not
               return until the task is complete. If False, it will spawn the import task and
               return immediately.
        """
        self.upload_file(file_id, content)
        return self.run_action(action_id, wait_for_completion)

    def export_and_download(self, action_id: int) -> bytes:
        """
        Convenience wrapper around `run_action()` and `get_file()` to run an export action and
        download the exported content in one call.
        :param action_id: The identifier of the action to run.
        :return: The content of the exported file.
        """
        self.run_action(action_id)
        return self.get_file(action_id)

    def get_task_summaries(self, action_id: int) -> list[TaskSummary]:
        """
        Retrieves the status of all tasks spawned by the specified action.
        :param action_id: The identifier of the action that was invoked.
        :return: The list of tasks spawned by the action.
        """
        return [
            TaskSummary.model_validate(e)
            for e in self._http.get_paginated(
                f"{self._url}/{action_url(action_id)}/{action_id}/tasks", "tasks"
            )
        ]

    def get_task_status(self, action_id: int, task_id: str) -> TaskStatus:
        """
        Retrieves the status of the specified task.
        :param action_id: The identifier of the action that was invoked.
        :param task_id: The identifier of the spawned task.
        :return: The status of the task.
        """
        return TaskStatus.model_validate(
            self._http.get(f"{self._url}/{action_url(action_id)}/{action_id}/tasks/{task_id}").get(
                "task"
            )
        )

    def get_optimizer_log(self, action_id: int, task_id: str) -> bytes:
        """
        Retrieves the solution logs of the specified optimization action task.
        :param action_id: The identifier of the optimization action that was invoked.
        :param task_id: The Task identifier, sometimes also referred to as the Correlation Id.
        :return: The content of the solution logs.
        """
        return self._http.get_binary(
            f"{self._url}/optimizeActions/{action_id}/tasks/{task_id}/solutionLogs"
        )

    def _file_pre_check(self, file_id: int) -> int:
        file = next((f for f in self.get_files() if f.id == file_id), None)
        if not file:
            raise InvalidIdentifierException(f"File {file_id} not found.")
        return file.chunk_count

    def _upload_chunk(self, file_id: int, index: int, chunk: str | bytes) -> None:
        self._http.put_binary_gzip(f"{self._url}/files/{file_id}/chunks/{index}", chunk)
        logger.debug(f"Chunk {index} loaded to file '{file_id}'.")

    def _set_chunk_count(self, file_id: int, num_chunks: int) -> None:
        logger.debug(f"Setting chunk count for file '{file_id}' to {num_chunks}.")
        if not self.allow_file_creation and not (113000000000 <= file_id <= 113999999999):
            raise InvalidIdentifierException(
                f"File with Id {file_id} does not exist. If you want to dynamically create files "
                "to avoid this error, set `allow_file_creation=True` on the calling instance. "
                "Make sure you have understood the implications of this before doing so. "
            )
        response = self._http.post(f"{self._url}/files/{file_id}", json={"chunkCount": num_chunks})
        optionally_new_file = int(response.get("file").get("id"))
        if optionally_new_file != file_id:
            if self.allow_file_creation:
                logger.info(f"Created new file with name '{file_id}', Id is {optionally_new_file}.")
                return
            raise InvalidIdentifierException(
                f"File with Id {file_id} did not exist and was created in Anaplan. You may want to "
                f"ask a model builder to remove it. If you want to dynamically create files "
                "to avoid this error, set `allow_file_creation=True` on the calling instance. "
                "Make sure you have understood the implications of this before doing so."
            )



from pydantic import BaseModel, ConfigDict, field_serializer
from pydantic.alias_generators import to_camel


class AnaplanModel(BaseModel):
    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True)

    @field_serializer(
        "action_id", "file_id", "process_id", "id", when_used="unless-none", check_fields=False
    )  # While these are of type int, they are serialized as strings in the API payloads
    def str_serializer(self, v: int) -> str:
        return str(v)



from datetime import datetime
from typing import Literal

from pydantic import Field

from ._base import AnaplanModel
from .cloud_works import _BaseIntegration


class FlowSummary(_BaseIntegration):
    id: str = Field(description="The unique identifier of this flow.")
    steps_count: int = Field(description="The number of steps in this flow.")


class ExceptionBehavior(AnaplanModel):
    type: Literal["failure", "partial_success"] = Field(
        description="The type of exception that this behavior applies to."
    )
    strategy: Literal["stop", "continue"] = Field(
        description="The strategy to handle the exception."
    )


class FlowStep(AnaplanModel):
    referrer: str = Field(
        description="The unique identifier of the referenced step or integration."
    )
    name: str = Field(description="The name of this flow step.")
    type: Literal["Process", "Import", "Export"] = Field(description="The type of this flow step.")
    created_by: str = Field(description="The user who created this step.")
    created_date: datetime = Field(description="The initial creation date of this step.")
    modified_date: datetime = Field(description="The last modification date of this step.")
    modified_by: str | None = Field(
        default=None, description="The user who last modified this step."
    )
    model_id: str = Field(description="The ID of the model this step belongs to.")
    workspace_id: str = Field(description="The ID of the workspace this step belongs to.")
    depends_on: list[str] | None = Field(
        default=[], description="The IDs of steps that this step depends on."
    )
    is_skipped: bool = Field(description="Whether this step is skipped during execution.")
    exception_behavior: list[ExceptionBehavior] = Field(
        description="Configuration for handling exceptions during step execution."
    )


class Flow(FlowSummary):
    version: Literal["2.0"] = Field(default="2.0", description="The version of this flow.")
    nux_visible: bool = Field(description="Whether this integration is visible in the UI.")
    steps: list[FlowStep] = Field(default=[], description="The steps in this flow.")


class FlowStepInput(AnaplanModel):
    type: Literal["Integration"] = Field(
        default="Integration", description="The type of this flow step."
    )
    referrer: str = Field(
        description="The unique identifier of the referenced step or integration."
    )
    depends_on: list[str] | None = Field(
        default=None, description="The IDs of steps that this step depends on."
    )
    is_skipped: bool = Field(
        default=False, description="Whether this step is skipped during execution."
    )
    exception_behavior: list[ExceptionBehavior] = Field(
        default=[
            ExceptionBehavior(type="failure", strategy="stop"),
            ExceptionBehavior(type="partial_success", strategy="continue"),
        ],
        description=(
            "Configuration for handling exceptions during step execution. Defaults to stopping on "
            "Failure and continuing on Partial Success."
        ),
    )


class FlowInput(AnaplanModel):
    name: str = Field(description="The name of this flow.")
    version: Literal["2.0"] = Field(default="2.0", description="The version of this flow.")
    type: Literal["IntegrationFlow"] = Field(
        default="IntegrationFlow", description="The type of this flow."
    )
    steps: list[FlowStepInput] = Field(
        description="The steps in this flow.", min_length=2, max_length=100
    )



from asyncio import gather

from anaplan_sdk.models.cloud_works import (
    Connection,
    Integration,
    RunError,
    RunStatus,
    RunSummary,
    SingleIntegration,
)


async def test_list_connections(client):
    connections = await client.cw.get_connections()
    assert isinstance(connections, list)
    assert all(isinstance(c, Connection) for c in connections)


async def test_create_connection_pydantic(client, az_blob_connection, registry):
    con_id = await client.cw.create_connection(az_blob_connection)
    assert con_id is not None
    registry["connections"].append(con_id)


async def test_create_connection_dict(client, az_blob_connection_dict, registry):
    con_id = await client.cw.create_connection(az_blob_connection_dict)
    assert con_id is not None
    registry["connections"].append(con_id)


async def test_update_connection_pydantic(client, az_blob_connection, registry):
    await client.cw.update_connection(registry["connections"][0], az_blob_connection.body)


async def test_update_connection_dict(client, az_blob_connection_dict, registry):
    await client.cw.update_connection(registry["connections"][-1], az_blob_connection_dict["body"])


async def test_patch_connection(client, name, registry):
    await client.cw.patch_connection(registry["connections"][-1], {"name": name})


async def test_get_integration(client, registry, test_integration):
    assert isinstance(await client.cw.get_integration(test_integration), SingleIntegration)


async def test_list_integrations(client):
    integrations_asc = await client.cw.get_integrations()
    assert isinstance(integrations_asc, list)
    assert all(isinstance(i, Integration) for i in integrations_asc)


async def test_list_integrations_desc(client):
    integrations_desc = await client.cw.get_integrations(sort_by="name", descending=True)
    assert isinstance(integrations_desc, list)
    assert all(isinstance(i, Integration) for i in integrations_desc)


async def test_create_integration_pydantic(client, integration_pydantic, registry):
    integration_id = await client.cw.create_integration(integration_pydantic)
    assert integration_id is not None
    registry["integrations"].append(integration_id)


async def test_create_multi_step_integration_pydantic(
    client, multi_step_integration_pydantic, registry
):
    integration_id = await client.cw.create_integration(multi_step_integration_pydantic)
    assert integration_id is not None
    registry["integrations"].append(integration_id)


async def test_create_process_integration_pydantic(client, process_integration_pydantic, registry):
    integration_id = await client.cw.create_integration(process_integration_pydantic)
    assert integration_id is not None
    registry["integrations"].append(integration_id)


async def test_update_integration_pydantic(client, integration_pydantic, registry):
    await client.cw.update_integration(registry["integrations"][0], integration_pydantic)


async def test_update_multi_step_integration_pydantic(
    client, registry, multi_step_integration_pydantic
):
    await client.cw.update_integration(registry["integrations"][1], multi_step_integration_pydantic)


async def test_update_process_integration_pydantic(client, registry, process_integration_pydantic):
    await client.cw.update_integration(registry["integrations"][-1], process_integration_pydantic)


async def test_create_integration_dict(client, integration_dict, registry):
    integration_id = await client.cw.create_integration(integration_dict)
    assert integration_id is not None
    registry["integrations"].append(integration_id)


async def test_create_multi_step_integration_dicts(client, multi_step_integration_dict, registry):
    integration_id = await client.cw.create_integration(multi_step_integration_dict)
    assert integration_id is not None
    registry["integrations"].append(integration_id)


async def test_create_process_integration_dicts(client, multi_step_integration_dict, registry):
    integration_id = await client.cw.create_integration(multi_step_integration_dict)
    assert integration_id is not None
    registry["integrations"].append(integration_id)


async def test_update_integration_dict(client, integration_dict, registry):
    await client.cw.update_integration(registry["integrations"][-3], integration_dict)


async def test_update_multi_step_integration_dict(client, registry, multi_step_integration_dict):
    await client.cw.update_integration(registry["integrations"][-2], multi_step_integration_dict)


async def test_update_process_integration_dict(client, registry, process_integration_pydantic):
    await client.cw.update_integration(registry["integrations"][-1], process_integration_pydantic)


async def test_run_integration(client, test_integration, registry):
    run_id = await client.cw.run_integration(test_integration)
    assert run_id is not None
    registry["run_id"] = run_id


async def test_get_run_history(client, registry):
    history = await client.cw.get_run_history(registry["integrations"][-1])
    assert isinstance(history, list)
    assert all((isinstance(i, RunSummary) for i in history))


async def test_get_run_status(client, registry):
    status = await client.cw.get_run_status(registry["run_id"])
    assert isinstance(status, RunStatus)


async def test_create_schedule_pydantic(client, registry, schedule_pydantic):
    await client.cw.create_schedule(registry["integrations"][0], schedule_pydantic)


async def test_update_schedule_pydantic(client, registry, schedule_pydantic):
    await client.cw.update_schedule(registry["integrations"][0], schedule_pydantic)


async def test_set_schedule_enabled(client, registry):
    await client.cw.set_schedule_status(registry["integrations"][0], "enabled")


async def test_set_schedule_disabled(client, registry):
    await client.cw.set_schedule_status(registry["integrations"][0], "disabled")


async def test_delete_schedule(client, registry):
    await client.cw.delete_schedule(registry["integrations"][0])


async def test_create_schedule_dict(client, registry, schedule_dict):
    await client.cw.create_schedule(registry["integrations"][0], schedule_dict)


async def test_update_schedule_dict(client, registry, schedule_dict):
    await client.cw.update_schedule(registry["integrations"][0], schedule_dict)


async def test_update_notification_dict(
    client, test_notification, notification_dict, test_integration
):
    notification_dict["integrationIds"] = [test_integration]
    await client.cw.update_notification_config(test_notification, notification_dict)


async def test_delete_notification(client, registry):
    await client.cw.delete_notification_config(integration_id=registry["integrations"][0])


async def test_create_notification_dict(client, notification_dict, registry):
    notification_dict["integrationIds"] = [registry["integrations"][0]]
    await client.cw.create_notification_config(notification_dict)


async def test_get_run_error(client, error_run_id):
    run_error = await client.cw.get_run_error(error_run_id)
    assert isinstance(run_error, RunError)


async def test_delete_integration(client, registry):
    await gather(*(client.cw.delete_integration(i) for i in registry["integrations"]))


async def test_delete_connection(client, registry):
    await gather(*(client.cw.delete_connection(c) for c in registry["connections"]))



from asyncio import gather

from anaplan_sdk.models.flows import Flow, FlowSummary


async def test_list_flows(client):
    flows = await client.cw.flows.get_flows()
    assert isinstance(flows, list)
    assert all(isinstance(f, FlowSummary) for f in flows)


async def test_list_flows_current_user(client):
    flows = await client.cw.flows.get_flows(current_user_only=True)
    assert isinstance(flows, list)
    assert all(isinstance(f, FlowSummary) for f in flows)


async def test_create_flow_pydantic(client, flow_pydantic, registry):
    flow_id = await client.cw.flows.create_flow(flow_pydantic)
    assert flow_id is not None
    registry["flows"].append(flow_id)


async def test_create_flow_dict(client, flow_dict, registry):
    flow_id = await client.cw.flows.create_flow(flow_dict)
    assert flow_id is not None
    registry["flows"].append(flow_id)


async def test_get_flow(client, registry):
    flow = await client.cw.flows.get_flow(registry["flows"][0])
    assert isinstance(flow, Flow)


async def test_run_flow(client, test_flow):
    run_id = await client.cw.flows.run_flow(test_flow)
    assert run_id is not None


async def test_update_flow_pydantic(client, flow_pydantic, registry):
    await client.cw.flows.update_flow(registry["flows"][0], flow_pydantic)


async def test_update_flow_dict(client, flow_dict, registry):
    await client.cw.flows.update_flow(registry["flows"][1], flow_dict)


async def test_delete_flow(client, registry):
    await gather(*(client.cw.flows.delete_flow(f) for f in registry["flows"]))



from ._alm import _AlmClient
from ._audit import _AuditClient
from ._bulk import Client
from ._cloud_works import _CloudWorksClient
from ._cw_flow import _FlowClient
from ._scim import _ScimClient
from ._transactional import _TransactionalClient

__all__ = [
    "Client",
    "_AlmClient",
    "_AuditClient",
    "_CloudWorksClient",
    "_FlowClient",
    "_ScimClient",
    "_TransactionalClient",
]



from os import getenv

import pytest

from anaplan_sdk import AsyncClient


@pytest.fixture(scope="session")
def client() -> AsyncClient:
    return AsyncClient(
        workspace_id=getenv("ANAPLAN_SDK_TEST_WORKSPACE_ID"),
        model_id=getenv("ANAPLAN_SDK_TEST_MODEL_ID"),
        certificate=getenv("ANAPLAN_SDK_TEST_CERT"),
        private_key=getenv("ANAPLAN_SDK_TEST_PK"),
        retry_count=3,
        backoff=5,
        timeout=120,
    )


@pytest.fixture(scope="session")
def client_small_pages() -> AsyncClient:
    return AsyncClient(
        workspace_id=getenv("ANAPLAN_SDK_TEST_WORKSPACE_ID"),
        model_id=getenv("ANAPLAN_SDK_TEST_MODEL_ID"),
        certificate=getenv("ANAPLAN_SDK_TEST_CERT"),
        private_key=getenv("ANAPLAN_SDK_TEST_PK"),
        page_size=100,
        retry_count=3,
        backoff=5,
        timeout=120,
    )


@pytest.fixture(scope="session")
def alm_src_client(client: AsyncClient, alm_src_model_id) -> AsyncClient:
    return client.with_model(alm_src_model_id)


@pytest.fixture(scope="session")
def alm_client(client: AsyncClient, alm_model_id) -> AsyncClient:
    return client.with_model(alm_model_id)


@pytest.fixture(scope="session")
def broken_client():
    return AsyncClient(
        workspace_id="",
        model_id="",
        certificate=getenv("ANAPLAN_SDK_TEST_CERT"),
        private_key=getenv("ANAPLAN_SDK_TEST_PK"),
        retry_count=1,
    )


@pytest.fixture(scope="session")
def test_list(py_version):
    if "3.10" in py_version:
        return 101000000309
    if "3.11" in py_version:
        return 101000000310
    if "3.12" in py_version:
        return 101000000311
    return 101000000312


@pytest.fixture(scope="session")
def test_file(py_version):
    if "3.10" in py_version:
        return 113000000061
    if "3.11" in py_version:
        return 113000000062
    if "3.12" in py_version:
        return 113000000063
    return 113000000064


@pytest.fixture(scope="session")
def test_action(py_version):
    if "3.10" in py_version:
        return 118000000028
    if "3.11" in py_version:
        return 118000000027
    if "3.12" in py_version:
        return 118000000026
    return 118000000025


@pytest.fixture(scope="session")
def alm_src_model_id(py_version):
    if "3.10" in py_version:
        return "C852660C6AF547258FB6B97D4942AB27"
    if "3.11" in py_version:
        return "C042E6A3F0334DEC9C826E1E12947789"
    if "3.12" in py_version:
        return "E130519009E54304A54D8D9F1C6D2725"
    return "12A4D6B9A816481F83F795F857407049"


@pytest.fixture(scope="session")
def alm_model_id(py_version):
    if "3.10" in py_version:
        return "174F0F20D5A84FF09C812E95B8E50997"
    if "3.11" in py_version:
        return "FFBB7D26E61040EBB5CACD80F7F8A01B"
    if "3.12" in py_version:
        return "FDB11C592E9445D78337F01BABC89880"
    return "0CBB17195E5D445A9A3692F481D66CBE"


@pytest.fixture(scope="session")
def test_integration(py_version):
    if "3.10" in py_version:
        return "840ccd8a279a454d99577d9538f24f09"
    if "3.11" in py_version:
        return "c0fa795faac047468a59c8dbe3752d75"
    if "3.12" in py_version:
        return "0204ea3261c8431e9e36ff1239c16247"
    return "cf9e1cf27a0f4eddb37a2a4807fd0ffc"


@pytest.fixture(scope="session")
def test_notification(py_version):
    if "3.10" in py_version:
        return "bfe29c0ff7434bde96c94ce1ec1b8e0a"
    if "3.11" in py_version:
        return "e2c709c74998460c8688b641cde07cd3"
    if "3.12" in py_version:
        return "e0f3d33a9c114e3a9c0e0908cffdb5e3"
    return "e57b5620f006444a9324baaa4bd891ff"


@pytest.fixture(scope="session")
def test_flow(py_version):
    if "3.10" in py_version:
        return "35e19e2f0f594d589f07fd8ba98c30a8"
    if "3.11" in py_version:
        return "0ca27f18a3f04a1382ecd1745609329b"
    if "3.12" in py_version:
        return "c9fd9841222d43d9886758ba4db4c340"
    return "c330ca2eda974650bd99fea50b0e3acd"


@pytest.fixture(scope="session")
def registry():
    return {"connections": [], "integrations": [], "flows": [], "run_id": None}



from typing import Any, Literal

from anaplan_sdk._services import _HttpService
from anaplan_sdk._utils import sort_params
from anaplan_sdk.models import User

Event = Literal["all", "byok", "user_activity"]
UserSortBy = Literal["first_name", "last_name", "email", "active", "last_login_date"] | None


class _AuditClient:
    def __init__(self, http: _HttpService) -> None:
        self._http = http
        self._limit = 10_000
        self._url = "https://audit.anaplan.com/audit/api/1/events"

    def get_users(
        self,
        search_pattern: str | None = None,
        sort_by: UserSortBy = None,
        descending: bool = False,
    ) -> list[User]:
        """
        Lists all the Users in the authenticated users default tenant.
        :param search_pattern: **Caution: This is an undocumented Feature and may behave
               unpredictably. It requires the Tenant Admin role. For non-admin users, it is
               ignored.** Optionally filter for specific users. When provided,
               case-insensitive matches users with emails or names containing this string.
               You can use the wildcards `%` for 0-n characters, and `_` for exactly 1 character.
               When None (default), returns all users.
        :param sort_by: The field to sort the results by.
        :param descending: If True, the results will be sorted in descending order.
        :return: The List of Users.
        """
        params = sort_params(sort_by, descending)
        if search_pattern:
            params["s"] = search_pattern
        res = self._http.get_paginated("https://api.anaplan.com/2/0/users", "users", params=params)
        return [User.model_validate(e) for e in res]

    def get_user(self, user_id: str = "me") -> User:
        """
        Retrieves information about the specified user, or the authenticated user if none specified.
        :return: The requested or currently authenticated User.
        """
        return User.model_validate(
            self._http.get(f"https://api.anaplan.com/2/0/users/{user_id}").get("user")
        )

    def get_events(
        self, days_into_past: int = 30, event_type: Event = "all"
    ) -> list[dict[str, Any]]:
        """
        Get audit events from Anaplan Audit API.
        :param days_into_past: The nuber of days into the past to get events for. The API provides
               data for up to 30 days.
        :param event_type: The type of events to get.
        :return: A list of log entries, each containing a dictionary with event details.
        """
        return list(
            self._http.get_paginated(
                self._url,
                "response",
                params={"type": event_type, "intervalInHours": days_into_past * 24},
            )
        )



import logging
from asyncio import gather
from itertools import chain
from typing import Any

from anaplan_sdk._services import _AsyncHttpService
from anaplan_sdk._utils import construct_payload
from anaplan_sdk.models.scim import (
    Operation,
    ReplaceUserInput,
    Resource,
    Schema,
    ServiceProviderConfig,
    User,
    UserInput,
    field,
)

logger = logging.getLogger("anaplan_sdk")


class _AsyncScimClient:
    def __init__(self, http: _AsyncHttpService) -> None:
        self._http = http
        self._url = "https://api.anaplan.com/scim/1/0/v2"

    async def get_service_provider_config(self) -> ServiceProviderConfig:
        """
        Get the SCIM Service Provider Configuration.
        :return: The ServiceProviderConfig object describing the available SCIM features.
        """
        res = await self._http.get(f"{self._url}/ServiceProviderConfig")
        return ServiceProviderConfig.model_validate(res)

    async def get_resource_types(self) -> list[Resource]:
        """
        Get the SCIM Resource Types.
        :return: A list of Resource objects describing the SCIM resource types.
        """
        res = await self._http.get(f"{self._url}/ResourceTypes")
        return [Resource.model_validate(e) for e in res.get("Resources", [])]

    async def get_resource_schemas(self) -> list[Schema]:
        """
        Get the SCIM Resource Schemas.
        :return: A list of Schema objects describing the SCIM resource schemas.
        """
        res = await self._http.get(f"{self._url}/Schemas")
        return [Schema.model_validate(e) for e in res.get("Resources", [])]

    async def get_users(self, predicate: str | field = None, page_size: int = 100) -> list[User]:
        """
        Get a list of users, optionally filtered by a predicate. Keep in mind that this will only
        return internal users. To get a list of all users in the tenant, use the `get_users()`
        in the `audit` namespace instead.
        :param predicate: A filter predicate to filter the users. This can either be a string,
               in which case it will be passed as-is, or an expression. Anaplan supports filtering
               on the following fields: "id", "externalId", "userName", "name.familyName",
               "name.givenName" and "active". It supports the operators "eq", "ne", "gt", "ge",
               "lt", "le" and "pr". It supports logical operators "and" and "or", "not" is not
               supported. It supports grouping with parentheses.
        :param page_size: The number of users to fetch per page. Values above 100 will error.
        :return: The internal users optionally matching the filter.
        """
        params: dict[str, int | str] = {"startIndex": 1, "count": page_size}
        if predicate is not None:
            _predicate = predicate if isinstance(predicate, str) else str(predicate)
            logger.debug(f"Searching for users with predicate: {_predicate}")
            params["filter"] = _predicate
        res = await self._http.get(f"{self._url}/Users", params=params)
        users = [User.model_validate(e) for e in res.get("Resources", [])]
        if (total := res["totalResults"]) <= page_size:
            return users
        pages = await gather(
            *(
                self._http.get(
                    f"{self._url}/Users", params=(params | {"startIndex": i, "count": page_size})
                )
                for i in range(page_size + 1, total + 1, page_size)
            )
        )
        for user in chain(*(p.get("Resources", []) for p in pages)):
            users.append(User.model_validate(user))
        return users

    async def get_user(self, user_id: str) -> User:
        """
        Get a user by their ID.
        :param user_id: The ID of the user to fetch.
        :return: The User object.
        """
        res = await self._http.get(f"{self._url}/Users/{user_id}")
        return User.model_validate(res)

    async def add_user(self, user: UserInput | dict[str, Any]) -> User:
        """
        Add a new user to your Anaplan tenant.
        :param user: The user info to add. Can either be a UserInput object or a dict. If you pass
               a dict, it will be validated against the UserInput model before sending. If the info
               you provided is invalid or incomplete, this will raise a pydantic.ValidationError.
        :return: The created User object.
        """
        res = await self._http.post(f"{self._url}/Users", json=construct_payload(UserInput, user))
        user = User.model_validate(res)
        logger.info(f"Added user '{user.user_name}' with ID '{user.id}'.")
        return user

    async def replace_user(self, user_id: str, user: ReplaceUserInput | dict[str, Any]):
        """
        Replace an existing user with new information. Note that this will replace all fields of the
        :param user_id: ID of the user to replace.
        :param user: The new user info. Can either be a ReplaceUserInput object or a dict. If you
               pass a dict, it will be validated against the ReplaceUserInput model before sending.
               If the info you provided is invalid or incomplete, this will raise a
               pydantic.ValidationError.
        :return: The updated User object.
        """
        res = await self._http.put(
            f"{self._url}/Users/{user_id}", json=construct_payload(ReplaceUserInput, user)
        )
        user = User.model_validate(res)
        logger.info(f"Replaced user with ID '{user_id}' with '{user.user_name}'.")
        return user

    async def update_user(
        self, user_id: str, operations: list[Operation] | list[dict[str, Any]]
    ) -> User:
        """
        Update an existing user with a list of operations. This allows you to update only specific
        fields of the user without replacing the entire user.
        :param user_id: The ID of the user to update.
        :param operations: A list of operations to perform on the user. Each operation can either be
               an Operation object or a dict. If you pass a dict, it will be validated against
               the Operation model before sending. If the operation is invalid, this will raise a
               pydantic.ValidationError. You can also use the models Replace, Add and Remove which
               are subclasses of Operation and provide a more convenient way to create operations.
        :return: The updated User object.
        """
        res = await self._http.patch(
            f"{self._url}/Users/{user_id}",
            json={
                "schemas": ["urn:ietf:params:scim:api:messages:2.0:PatchOp"],
                "Operations": [construct_payload(Operation, e) for e in operations],
            },
        )
        user = User.model_validate(res)
        logger.info(f"Updated user with ID '{user_id}'.")
        return user



from calendar import monthrange
from datetime import date
from os import getenv

from anaplan_sdk import AsyncClient
from anaplan_sdk.models import (
    CurrentPeriod,
    Dimension,
    DimensionWithCode,
    FiscalYear,
    InsertionResult,
    ListItem,
    ListMetadata,
    Model,
    ModelStatus,
    MonthsQuartersYearsCalendar,
    View,
    ViewInfo,
)


async def test_wake_model(client: AsyncClient):
    await client.tr.wake_model()


async def test_close_model(client: AsyncClient):
    await client.with_model("C87EBE934BD442B1A798540E0CA5A877").tr.close_model()


async def test_get_model(client: AsyncClient):
    model_id = getenv("ANAPLAN_SDK_TEST_MODEL_ID")
    model = await client.tr.get_model_details()
    assert isinstance(model, Model)
    assert model.id == model_id


async def test_list_modules(client: AsyncClient):
    modules = await client.tr.get_modules()
    assert isinstance(modules, list)
    assert len(modules) > 0


async def test_list_lists(client: AsyncClient):
    lists = await client.tr.get_lists()
    assert isinstance(lists, list)
    assert len(lists) > 0


async def test_list_line_items(client: AsyncClient):
    items = await client.tr.get_line_items()
    assert isinstance(items, list)
    assert len(items) > 0


async def test_get_list_meta(client: AsyncClient, test_list):
    meta = await client.tr.get_list_metadata(test_list)
    assert isinstance(meta, ListMetadata)


async def test_get_model_status(client: AsyncClient):
    status = await client.tr.get_model_status()
    assert isinstance(status, ModelStatus)


async def test_long_list_insertion(client: AsyncClient, test_list, list_items_long):
    result = await client.tr.insert_list_items(test_list, list_items_long)
    assert isinstance(result, InsertionResult)
    assert result.failures == []
    assert result.added == 200_000
    assert result.total == 200_000


async def test_long_list_deletion(client: AsyncClient, test_list, list_items_long):
    result = await client.tr.delete_list_items(test_list, list_items_long)
    assert result.deleted == 200_000
    assert result.failures == []


async def test_short_list_insertion(client: AsyncClient, test_list, list_items_short):
    result = await client.tr.insert_list_items(test_list, list_items_short)
    assert isinstance(result, InsertionResult)
    assert result.failures == []
    assert result.added == 1_000
    assert result.total == 1_000


async def test_get_list_items(client: AsyncClient, test_list):
    items = await client.tr.get_list_items(test_list, False)
    assert isinstance(items, list)
    assert len(items) == 1_000
    assert all(isinstance(item, ListItem) for item in items)


async def test_get_list_items_raw(client: AsyncClient, test_list):
    items = await client.tr.get_list_items(test_list, True)
    assert isinstance(items, list)
    assert len(items) == 1_000
    assert all(isinstance(item, dict) for item in items)


async def test_short_list_deletion(client: AsyncClient, test_list, list_items_short):
    result = await client.tr.delete_list_items(test_list, list_items_short)
    assert result.deleted == 1_000
    assert result.failures == []


async def test_reset_list_index(client: AsyncClient, test_list):
    await client.tr.reset_list_index(test_list)


async def test_list_views(client: AsyncClient):
    views = await client.tr.get_views()
    assert isinstance(views, list)
    assert len(views) > 0
    assert all(isinstance(view, View) for view in views)


async def test_get_view_info(client: AsyncClient):
    info = await client.tr.get_view_info(102000000015)
    assert isinstance(info, ViewInfo)


async def test_get_current_period(client: AsyncClient):
    period = await client.tr.get_current_period()
    assert isinstance(period, CurrentPeriod)


async def test_set_current_period(client: AsyncClient):
    today = date.today()
    last_day_of_month = date(today.year, today.month, monthrange(today.year, today.month)[1])
    period = await client.tr.set_current_period(today.strftime("%Y-%m-%d"))
    assert isinstance(period, CurrentPeriod)
    assert period.last_day == last_day_of_month.strftime("%Y-%m-%d")


async def test_set_current_fiscal_year(client: AsyncClient):
    year = "FY25"
    fiscal_year = await client.tr.set_current_fiscal_year(year)
    assert isinstance(fiscal_year, FiscalYear)
    assert fiscal_year.year == year


async def test_get_model_calendar(client: AsyncClient):
    calendar = await client.tr.get_model_calendar()
    assert isinstance(calendar, MonthsQuartersYearsCalendar)


async def test_get_dimension_items(client: AsyncClient):
    items = await client.tr.get_dimension_items(109000000000)
    assert isinstance(items, list)
    assert all(isinstance(item, DimensionWithCode) for item in items)


async def test_get_dimension_items_with_list_warns(client: AsyncClient, caplog):
    items = await client.tr.get_dimension_items(101000000008)
    assert isinstance(items, list)
    assert all(isinstance(item, DimensionWithCode) for item in items)
    assert any(
        ("warn" in record.levelname.lower() and "is discouraged." in record.msg)
        for record in caplog.records
    )


async def test_get_dimension_items_with_users_warns(client: AsyncClient, caplog):
    items = await client.tr.get_dimension_items(101999999999)
    assert isinstance(items, list)
    assert all(isinstance(item, DimensionWithCode) for item in items)
    assert any(
        ("warn" in record.levelname.lower() and "is discouraged." in record.msg)
        for record in caplog.records
    )


async def test_get_line_item_dimensions(client: AsyncClient):
    items = await client.tr.get_line_item_dimensions(284000000077)
    assert isinstance(items, list)
    assert all(isinstance(item, Dimension) for item in items)



from os import getenv

import pytest

from anaplan_sdk import Client


@pytest.fixture(scope="session")
def client() -> Client:
    return Client(
        workspace_id=getenv("ANAPLAN_SDK_TEST_WORKSPACE_ID"),
        model_id=getenv("ANAPLAN_SDK_TEST_MODEL_ID"),
        certificate=getenv("ANAPLAN_SDK_TEST_CERT"),
        private_key=getenv("ANAPLAN_SDK_TEST_PK"),
        retry_count=3,
        backoff=5,
        timeout=120,
    )


@pytest.fixture(scope="session")
def client_small_pages() -> Client:
    return Client(
        workspace_id=getenv("ANAPLAN_SDK_TEST_WORKSPACE_ID"),
        model_id=getenv("ANAPLAN_SDK_TEST_MODEL_ID"),
        certificate=getenv("ANAPLAN_SDK_TEST_CERT"),
        private_key=getenv("ANAPLAN_SDK_TEST_PK"),
        page_size=100,
        retry_count=3,
        backoff=5,
        timeout=120,
    )


@pytest.fixture(scope="session")
def alm_src_client(client: Client, alm_src_model_id) -> Client:
    return client.with_model(alm_src_model_id)


@pytest.fixture(scope="session")
def alm_client(client: Client, alm_model_id) -> Client:
    return client.with_model(alm_model_id)


@pytest.fixture(scope="session")
def broken_client() -> Client:
    return Client(
        workspace_id="",
        model_id="",
        certificate=getenv("ANAPLAN_SDK_TEST_CERT"),
        private_key=getenv("ANAPLAN_SDK_TEST_PK"),
        retry_count=1,
    )


@pytest.fixture(scope="session")
def test_list(py_version):
    if "3.10" in py_version:
        return 101000000305
    if "3.11" in py_version:
        return 101000000306
    if "3.12" in py_version:
        return 101000000307
    return 101000000308


@pytest.fixture(scope="session")
def test_file(py_version):
    if "3.10" in py_version:
        return 113000000065
    if "3.11" in py_version:
        return 113000000066
    if "3.12" in py_version:
        return 113000000067
    return 113000000068


@pytest.fixture(scope="session")
def test_action(py_version):
    if "3.10" in py_version:
        return 118000000024
    if "3.11" in py_version:
        return 118000000023
    if "3.12" in py_version:
        return 118000000022
    return 118000000021


@pytest.fixture(scope="session")
def alm_src_model_id(py_version):
    if "3.10" in py_version:
        return "662239743F56420EBECAE7EE0659475C"
    if "3.11" in py_version:
        return "4FCB1B5224F64339AADC7B322E1663EF"
    if "3.12" in py_version:
        return "A68B26D67D00443FB8A3428CC83E427D"
    return "5327704441464F36B4B05C6BF9DFACD2"


@pytest.fixture(scope="session")
def alm_model_id(py_version):
    if "3.10" in py_version:
        return "7538F3BE46B94F208C6AF5051919E56E"
    if "3.11" in py_version:
        return "5C31E832E0B04567B5A8784B25AF3572"
    if "3.12" in py_version:
        return "7F51D466D915425CA80BEDF54BF364AE"
    return "33A804FE8FE34AC291A6E866CA6B2284"


@pytest.fixture(scope="session")
def test_integration(py_version):
    if "3.10" in py_version:
        return "44bd0bd4606b4f77b62e70d5ff617f3f"
    if "3.11" in py_version:
        return "dfcf1caace4f41748fc589e83c68c65a"
    if "3.12" in py_version:
        return "18e2c03b0bdf4593bdd964786891ead8"
    return "c5bd6fd4b9414ea0959795ccebc8126a"


@pytest.fixture(scope="session")
def test_notification(py_version):
    if "3.10" in py_version:
        return "efc8e3340c054f00bc20dbab1719531f"
    if "3.11" in py_version:
        return "81f0e8c718ed47ada55ce9b88dad548f"
    if "3.12" in py_version:
        return "d90749ebb0e94692b4eaba1963cf0959"
    return "393d6e91e1874e28baccd7441d28ba36"


@pytest.fixture(scope="session")
def test_flow(py_version):
    if "3.10" in py_version:
        return "8f9a377127844984b12775e9ca072108"
    if "3.11" in py_version:
        return "240a3d55754843be824cb4f97e48e061"
    if "3.12" in py_version:
        return "afc6cec5d4b343c6abb4f0b3d41f1f7f"
    return "61f2a2b05ac64598bfeb6d9eb7eff97a"


@pytest.fixture(scope="session")
def registry():
    return {"connections": [], "integrations": [], "flows": [], "run_id": None}
